local SKIN = "jetmetalsonic"

//Values for Charge Abilities
local MAX_CHARGES = 2		--Maximum amount of charges per jump
local SUPER_MAX_CHARGES = 4	--Maximum amount of charges per jump when super

local WATERFACTOR = 3*FRACUNIT/5	--Factor for adjustment when underwater
local SUPERFACTOR = 6*FRACUNIT/5	--Factor for adjustment when Super

local FULL_CHARGE_BUILDUP = 17*TICRATE/15			--Amount of charge at maximum, set in global table for HUD script
local HALF_CHARGE_BUILDUP = FULL_CHARGE_BUILDUP/2	--Amount of charge at half-full
local MAX_CHARGE_TIME = 2*TICRATE					--Maximum amount of time allowed to be charging

local BASE_BUILDUP_MOD = 5*FRACUNIT/4	--Buildup modifier at minimum
local ADD_BUILDUP_MOD = 6*FRACUNIT/5	--Buildup modifier at full charge

local BASE_CHRGJMP_HEIGHT = 15*FRACUNIT/2	--Base value for Charge Jump height

local BASE_CHRGTHOK_SPEED = 20*FRACUNIT		--Base value for Charge Thok speed
local WALLBOUNCE_WINDOW = 4*TICRATE/5		--Window of opportunity to perform the wall bounce
local WALLBOUNCEXY_FACTOR = 4*FRACUNIT/11	--Modifier for horizontal bounce momentum
local WALLBOUNCEZ_FACTOR = 12*FRACUNIT		--Modifier for vertical bounce momentum

local BASE_V_STMPSPEED = -4*FRACUNIT		--Base value for vertical Stomp speed
local BASE_H_STMPSPEED = 6*FRACUNIT			--Base value for horizontal Stomp speed
local BASE_GBOUNCE_HEIGHT = 17*FRACUNIT		--Base value for Ground Bounce height

//Values for Homing Thok
local SEARCH_DIST = 6*RING_DIST/5	--Search distance for valid homing attack targets
local SEARCH_ANGLE = ANGLE_45		--Angle within which to search for valid homing targets

local SPEED_MOD = 2*FRACUNIT/3		--Modifier for movement speed during homing attack
local HOME_TIMER = 3*TICRATE		--Homing attack duration

//Values for Dash Mode
local DASHMODE_ON = 3*TICRATE			--Required build-up to activate Dash Mode
local DASHMODE_CAP = 11*TICRATE/2		--Maximum Dash Mode build-up
local DASHSPD_MOD_CAP = 4*FRACUNIT/3	--Maximum normalspeed multiplier in Dash Mode
local DASHMODE_DECAY = 3*TICRATE/4		--Speed at which Dash Mode decays

//Functions for visual effects
local function SpawnChargeEffect(p, color)
	local effect = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_CHRGEFF)
	effect.color = color
	effect.scale = p.mo.scale/2
	effect.tracer = p.mo
end

local function SpawnChargeEffect2(mo, color, freq)
	if leveltime % freq ~= 0 return end
	local x_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), 8*mo.radius)
	local y_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), 8*mo.radius)
	local z_adjust

	if P_IsObjectOnGround(mo)
		z_adjust = FixedMul(FixedMul(P_RandomFixed(), 16*mo.scale/5), mo.height)
		if mo.eflags & MFE_VERTICALFLIP
			z_adjust = mo.height - mo.height / 3 - FixedMul(FixedMul(P_RandomFixed(), 16*mo.scale/5), mo.height)
		end
	else
		z_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), 16*mo.scale/5), mo.height)
		if mo.eflags & MFE_VERTICALFLIP
			z_adjust = mo.height - mo.height / 3 - FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), 16*mo.scale/5), mo.height)
		end
	end

	local effect = P_SpawnMobj(mo.x + x_adjust, mo.y + y_adjust, mo.z + z_adjust, MT_2NDCHRGEFF)
	effect.scale = mo.scale/18
	effect.color = color
	if mo.eflags & MFE_VERTICALFLIP
		effect.eflags = $|MFE_VERTICALFLIP
	end
	effect.tracer = mo
end

local SpawnLaunchEffect = function(p, abil, size, color)
	local effect = P_SpawnMobj(p.mo.x, p.mo.y, (p.mo.z + 36 * p.mo.scale), MT_LAUNCHEFF)
	effect.color = color
	effect.scale = FixedMul(size, p.mo.scale)
	if abil == 1
		effect.frame = $|FF_PAPERSPRITE
		effect.angle = p.mo.angle + ANGLE_90
	else
		effect.frame = $|FF_FLOORSPRITE
	end
end

local function BoosterJetFume(mo, ghost)
	local x_adjust = P_ReturnThrustX(mo, mo.angle, -17*mo.scale)
	local y_adjust = P_ReturnThrustY(mo, mo.angle, -17*mo.scale)
	if not (mo.eflags & MFE_UNDERWATER)
		local jetfume = P_SpawnMobjFromMobj(mo, x_adjust, y_adjust, 0, MT_THOK)
		jetfume.fuse = 1
		jetfume.sprite = SPR_JETF
		jetfume.colorized = true
		jetfume.color = mo.target.player.effcolor
		jetfume.scale = mo.scale/2

		if leveltime % 3 == 0
			jetfume.frame = A|FF_FULLBRIGHT
		elseif leveltime % 3 == 1
			jetfume.frame = B|FF_FULLBRIGHT
		else
			jetfume.frame = C|FF_FULLBRIGHT
		end

		if mo.eflags & MFE_VERTICALFLIP
			jetfume.eflags = $|MFE_VERTICALFLIP
			P_TeleportMove(jetfume, jetfume.x, jetfume.y, jetfume.z + FixedMul(32*FRACUNIT, mo.scale))
		end

		if ghost
			P_SpawnGhostMobj(jetfume)
		end

	else
		local bubble = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SMALLBUBBLE) --Spawn bubbles instead when underwater
		if mo.eflags & MFE_VERTICALFLIP
			P_TeleportMove(bubble, mo.x + x_adjust, mo.y + y_adjust, mo.z + mo.scale)
			bubble.eflags = $|MFE_VERTICALFLIP
		else
			P_TeleportMove(bubble, mo.x + x_adjust, mo.y + y_adjust, mo.z - mo.scale)
		end
	end
end

local function JetScreamEffect(mo, color, freq, fuse, objecttype)
	if leveltime % freq ~= 0 return end
	local x_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), mo.radius)
	local y_adjust = FixedMul(FixedMul(P_RandomRange(-FRACUNIT/2, FRACUNIT/2 - 1), mo.scale), mo.radius)
	local z_adjust = FixedMul(FixedMul(P_RandomFixed(), mo.scale), mo.height)
	if mo.eflags & MFE_VERTICALFLIP
		z_adjust = mo.height - mo.height / 3 - FixedMul(FixedMul(P_RandomFixed(), mo.scale), mo.height)
	end

	local effect = P_SpawnMobj(mo.x + x_adjust, mo.y + y_adjust, mo.z + z_adjust, objecttype)
	effect.fuse = fuse
	if color ~= nil
		effect.color = color
	end
	if mo.eflags & MFE_VERTICALFLIP
		effect.eflags = $|MFE_VERTICALFLIP
	end

	if not mo.player return end
	if mo.player.powers[pw_super]
		P_InstaThrust(effect, R_PointToAngle2(mo.x, mo.y, effect.x, effect.y), FixedMul(3*FRACUNIT, mo.scale))
		if mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6
			effect.momx = 6*$
			effect.momy = 6*$
		end
		P_SetObjectMomZ(effect, 7*FRACUNIT/2, false)
	end
end

local function WallBounceDust(mo) --Dust from Krabs's Rebound Dash
	for i = 0, 9
		local dust = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale*18, MT_SPINDUST)
		dust.momx = (mo.momx*2)+(P_RandomRange(-10, 10)*FRACUNIT)
		dust.momy = (mo.momy*2)+(P_RandomRange(-10, 10)*FRACUNIT)
		dust.momz = P_RandomRange(-3, 3)*FRACUNIT
		dust.scale = P_RandomRange(50, 125)*FRACUNIT/100
		dust.destscale = 0
		dust.state = S_BOSSEXPLODE
	end
end

local function SpawnLightCircle(p, color)
	for i = 0, 8
		local fa = (i * (ANGLE_180 / 4))
		local radius = FixedMul(20*FRACUNIT, p.mo.scale)

		local offsetx = FixedMul(cos(fa), radius)
		local offsety = FixedMul(sin(fa), radius)

		local lightning = P_SpawnMobjFromMobj(p.mo, offsetx, offsety, 0, MT_LIGHTCIRCLE)
		lightning.target = p.mo
		lightning.angle = fa + ANGLE_90
		lightning.color = color
		lightning.circleradius = 20*FRACUNIT
	end
end

//Functions for Charge Abilities
local ResetCharge = function(p)
	p.chrgbuildUp = 0
	p.charging = false
	p.chargepercent = 0
	p.buildupmod = 0
	p.effsize = 0
	p.halfchargeeff = false
	p.fullchargeeff = false
end

local function ResetAbilities(p)
	p.chrgjumpUsed = false
	p.superchrgjumpUsed = false
	if p.chrgStomping
		p.prevChrgStomp = true
	end
	p.chrgStomping = false
	p.prepchrgBounce = false
	p.chrgWBounce = 0
	p.chrgGBounce = false
end

local function SuperNuke(p, radius) --Destroy nearby enemies if super
	if p.powers[pw_super]
		P_StartQuake(20*FRACUNIT, 4)
		P_NukeEnemies(p.mo, p.mo, radius)
		S_StartSound(p.mo, sfx_bkpoof)
	end
end

local function ConvertCharge(p) --Function to convert charge into a usable variable and set effect size
	p.effsize = 3*FRACUNIT/5
	if p.chrgbuildUp >= FULL_CHARGE_BUILDUP
		p.chrgbuildUp = FULL_CHARGE_BUILDUP
		p.effsize = FRACUNIT
		SuperNuke(p, 475*FRACUNIT)
	elseif p.chrgbuildUp > HALF_CHARGE_BUILDUP
		p.effsize = 4*FRACUNIT/5
	end
	p.buildupmod = BASE_BUILDUP_MOD+FixedMul(ADD_BUILDUP_MOD, FixedDiv(p.chrgbuildUp*FRACUNIT, FULL_CHARGE_BUILDUP*FRACUNIT))
end

local function FloorBreak(mo, sec) --Function to break breakable floors, adapted from MAXControl by Inferno
	for x = mo.x, mo.x + mo.radius, mo.radius
		for y = mo.y, mo.y + mo.radius, mo.radius
			for fof in sec.ffloors()
				if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP)
				and (mo.z + mo.height == fof.bottomheight or mo.z == fof.topheight
				or (mo.eflags & MFE_VERTICALFLIP) and (mo.z - mo.height == fof.topheight or mo.z == fof.bottomheight))
					EV_CrumbleChain(fof)
				end
			end
		end
	end
end

local function doChargeJump(p)
	local mo = p.mo

	local JUMPHEIGHT = FixedMul(p.buildupmod, BASE_CHRGJMP_HEIGHT) --Calculate Jump Height
	if p.powers[pw_super]
		JUMPHEIGHT = FixedMul($, SUPERFACTOR) --Increase jump height when Super
	end
	if mo.eflags & MFE_UNDERWATER
		JUMPHEIGHT = FixedMul($, WATERFACTOR) --Reduce jump height when underwater
	end
	P_SetObjectMomZ(mo, JUMPHEIGHT, false)
	P_InstaThrust(mo, mo.angle, FixedMul(p.storeSpd/4, mo.scale))  --Give forward speed equal to 25% stored speed
	mo.state = S_PLAY_CHRGJMP
	p.ghostTrail = 14

	if not p.prepchrgBounce
		p.prepchrgBounce = false
	end
	if p.prepchrgBounce
		p.prepchrgBounce = false
	end

	p.pflags = $ & ~PF_NOJUMPDAMAGE

	--Lose a charge
	p.charges = $-1
	if p.chrgjumpUsed
		p.superchrgjumpUsed = true
	end
	p.chrgjumpUsed = true
	ResetCharge(p)
end

local function doChargeThok(p)
	local mo = p.mo

	local THOKSPEED = (2*p.storeSpd/3)+FixedMul(p.buildupmod, BASE_CHRGTHOK_SPEED) --Calculate Thok speed
	if THOKSPEED > p.thokspeedcap
		THOKSPEED = p.thokspeedcap
	end
	if p.powers[pw_super]
		THOKSPEED = FixedMul($, SUPERFACTOR) --Increase thok speed when Super
	end
	if mo.eflags & MFE_UNDERWATER
		THOKSPEED = FixedMul($, WATERFACTOR)  --Reduce thok speed when underwater
	end
	P_InstaThrust(mo, mo.angle, FixedMul(THOKSPEED, mo.scale))
	if mo.eflags & MFE_UNDERWATER
		P_SetObjectMomZ(mo, 3*FRACUNIT, false)
	else
		P_SetObjectMomZ(mo, 5*FRACUNIT, false)
	end
	mo.state = S_PLAY_CHRGTHK

	if not p.prepchrgBounce
		p.prepchrgBounce = false
	end
	if p.prepchrgBounce
		p.chrgWBounce = WALLBOUNCE_WINDOW
		p.ghostTrail = WALLBOUNCE_WINDOW
		p.prepchrgBounce = false
	else
		p.ghostTrail = 14
	end

//	SuperNuke(p, 475*FRACUNIT)
	p.pflags = $ & ~PF_NOJUMPDAMAGE

	--Lose a charge
	p.charges = $-1
	ResetCharge(p)
end

local function doChargeStomp(p)
	local mo = p.mo

	p.STOMPSPEED = FixedMul(p.buildupmod, BASE_V_STMPSPEED)
	if p.powers[pw_super]
		p.STOMPSPEED = FixedMul($, SUPERFACTOR) --Increase stomp speed when Super
	end
	if mo.eflags & MFE_UNDERWATER
		p.STOMPSPEED = FixedMul($, WATERFACTOR)  --Reduce stomp speed when underwater
	end
	mo.momx = 0
	mo.momy = 0
	if p.effsize > 0
		P_InstaThrust(mo, mo.angle, FixedMul(p.storeSpd/2+BASE_H_STMPSPEED, mo.scale))
	end
	P_SetObjectMomZ(mo, p.STOMPSPEED, true)
	mo.state = S_PLAY_ROLL
	p.ghostTrail = 14

	if not p.prepchrgBounce
		p.prepchrgBounce = false
	end
	if p.prepchrgBounce
		p.chrgGBounce = true
		p.chrgGBouncing = true
		p.prepchrgBounce = false
	end

//	SuperNuke(p, 475*FRACUNIT)
	p.pflags = $ & ~PF_NOJUMPDAMAGE

	--Lose a charge
	p.charges = $-1
	p.chrgStomping = true
	ResetCharge(p)
end

//Functions for Homing Thok
local function JetLookForEnemies(p, ringdist, span, above, disregard) --Enemy search. Code re-used and modified from CobaltBW's Fang Popgun Modifications
	local twod = (twodlevel or p.mo.flags2 & MF2_TWOD)
	if (twod)
		ringdist = $/2
		span = 2*$/3
	end

	local maxdist = FixedMul(ringdist, p.mo.scale)
	local closestdist = 0
	local closestmo = nil
	local nonenemiesdisregard = disregard
	searchBlockmap("objects",function(pmo,mo)
		if (mo.flags & MF_NOCLIPTHING) return end
		if (mo.health <= 0) return end

		if not(mo.p)
		and (!((mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)
		and (mo.flags & MF_SHOOTABLE)) or (mo.flags & MF_SPRING)) == !(mo.flags2 & MF2_INVERTAIMABLE))
			return
		end
		if (mo == pmo) return end
		if (mo.flags2 & MF2_FRET) return end
		if (mo.flags & nonenemiesdisregard) return end
		if mo.type == MT_DETON return end -- Do not target Detons
		if mo.type == MT_TNTBARREL return end -- Do not target TNT barrels
		if mo.type == MT_METALSONIC_BATTLE return end -- Do not target Metal Sonic (Boss 6)

		//Disallow targeting teammates
		if mo.p and (mo.p.spectator or gametyperules&GTR_FRIENDLY or G_GametypeHasTeams() and mo.p.ctfteam == p.ctfteam)
			return
		end

		//Do angle/distance checks
		local zdist = (pmo.z + pmo.height/2) - (mo.z + mo.height/2)
		local dist = P_AproxDistance(pmo.x - mo.x, pmo.y - mo.y)
		//CBW:	Made the angle checks their own locals, for readability purposes
		//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
		local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
		local xy_angle = abs(R_PointToAngle2(
				pmo.x + P_ReturnThrustX(pmo, pmo.angle, pmo.radius),
				pmo.y + P_ReturnThrustY(pmo, pmo.angle, pmo.radius),
				mo.x, mo.y) - pmo.angle)

		dist = P_AproxDistance(dist, zdist)
		if (dist > maxdist)
			return -- out of range
		end
		if (xyz_angle > span)
			return -- Don't home outside of desired angle!
		end

		if not above -- Don't home above the player if not desired!
			if pmo.eflags & MFE_VERTICALFLIP
			and zdist > 0
				return
			elseif zdist < 0
				return
			end
		end

		if (twod
		and abs(pmo.y-mo.y) > pmo.radius)
			return -- not in your 2d plane
		end

		if ((closestmo and closestmo.valid) and (dist > closestdist))
			return
		end
		if (xy_angle > span)
			return -- behind back
		end

		if not (P_CheckSight(pmo, mo))
			return -- out of sight
		end

		closestmo = mo
		closestdist = dist
	end,p.mo,p.mo.x-maxdist,p.mo.x+maxdist,p.mo.y-maxdist,p.mo.y+maxdist)
	return closestmo
end

local function JetHoming(mo, t, spdmod)
	mo.angle = R_PointToAngle2(mo.x, mo.y, t.x, t.y)

	local dist = P_AproxDistance(P_AproxDistance(t.x - mo.x, t.y - mo.y), t.z - mo.z)
	local speed = FixedMul(FixedMul(mo.player.actionspd, spdmod), mo.scale)
	if mo.eflags & MFE_UNDERWATER
		speed = FixedMul($, WATERFACTOR) --Reduce speed when underwater
	end

	mo.momx = FixedMul(FixedDiv(t.x - mo.x, dist), speed)
	mo.momy = FixedMul(FixedDiv(t.y - mo.y, dist), speed)
	mo.momz = FixedMul(FixedDiv(t.z - mo.z, dist), speed)
end

//Functions for Dash Mode
local function DashModeReset(p)
	if p.dashModeactivate
		S_StartSound(p.mo, sfx_kc65) --Vanilla Dash Mode deactivation sound
		p.dashModeactivate = nil
	end
	if not (skins[p.mo.skin].flags & SF_RUNONWATER)
		p.charflags = $ & ~SF_RUNONWATER
	end
	if p.normalspeed ~= skins[p.mo.skin].normalspeed
		p.normalspeed = skins[p.mo.skin].normalspeed
	end
	if p.jumpfactor ~= skins[p.mo.skin].jumpfactor
		p.jumpfactor = skins[p.mo.skin].jumpfactor
	end
end

//Custom object thinkers
function A_BoosterThink(actor, var1, var2) --Thinker for the boosters
	local t = actor.target
	local p = t.player

	actor.color = t.color
	actor.colorized = t.colorized
	actor.scale = t.scale

	if t.health <= 0
		if leveltime % 8 == 0
			A_BossScream(actor, 1, MT_BOSSEXPLODE)
			S_StartSound(mo, sfx_s3kb4)
		end
		actor.flags = $ & ~(MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY)

		if not p.deadalready
			P_InstaThrust(actor, actor.angle + P_RandomRange(-20, 20) * ANG1, p.prevspeed/2)
			actor.momz = 2*t.momz/3
			actor.fuse = 5*TICRATE/2
		elseif not P_IsObjectOnGround(actor)
			if actor.type == MT_JETMETBSTERA
				actor.rollangle = $ - ANG10
			else
				actor.rollangle = $ + ANG10
			end
		else
			actor.momx = 0
			actor.momy = 0
		end
		return
	end

	--Var1 sets vertical offset
	local offsetz
	if t.eflags & MFE_VERTICALFLIP
		actor.eflags = $|MFE_VERTICALFLIP
		actor.flags2 = $|MF2_OBJECTFLIP
		offsetz = t.height - actor.height - FixedMul(var1, t.scale)
	else
		actor.eflags = $ & ~MFE_VERTICALFLIP
		actor.flags2 = $ & ~MF2_OBJECTFLIP
		offsetz = FixedMul(var1, actor.scale)
	end
	if leveltime % (2*TICRATE) < TICRATE --Bob up and down
		offsetz = FixedMul($ + (leveltime % TICRATE) / 8*FRACUNIT, t.scale)
	else
		offsetz = FixedMul($ + 34*FRACUNIT/8 - (leveltime % TICRATE) / 8*FRACUNIT, t.scale)
	end

	local angle = p.drawangle
	actor.angle = angle

	--Var2 sets horizontal offsets
	local foffsetx = P_ReturnThrustX(t, angle, FixedMul((var2>>FRACBITS)*FRACUNIT, t.scale))
	local foffsety = P_ReturnThrustY(t, angle, FixedMul((var2>>FRACBITS)*FRACUNIT, t.scale))
	local boffsetx = P_ReturnThrustX(t, angle + ANGLE_90, FixedMul((var2 % FRACUNIT)*FRACUNIT, t.scale))
	local boffsety = P_ReturnThrustY(t, angle + ANGLE_90, FixedMul((var2 % FRACUNIT)*FRACUNIT, t.scale))

	--Invert horizontal position for the second booster
	if actor.type == MT_JETMETBSTERB
		boffsetx = -$
		boffsety = -$
	end

	P_TeleportMove(actor, t.x + foffsetx + boffsetx, t.y + foffsety + boffsety, t.z + offsetz)
end

function A_ChargeThink(actor, var1, var2) --Thinker for the primary charge effect
	A_CapeChase(actor, var1, var2)
	if var1 % FRACUNIT
		actor.angle = actor.tracer.angle + ANGLE_90
	else
		actor.angle = actor.target.angle + ANGLE_90
	end
end

function A_2ndChargeThink(actor, var1, var2) --Thinker for the secondary charge effect
	local t = actor.tracer
	actor.angle = R_PointToAngle2(actor.x, actor.y, t.x, t.y)

	if not actor.fuse
		actor.fuse = var1 --Var1 sets fuse length
	end

	local ghost = P_SpawnGhostMobj(actor)
	ghost.fuse = var2 --Var2 sets ghost fuse length
	ghost.scale = 4*actor.scale/5

	local x_adjust = 0
	local y_adjust = 0
	local z_adjust = t.height/2

	if t.state ~= S_PLAY_SPINDASH
		x_adjust = P_ReturnThrustX(t, t.angle, FixedMul(-12*FRACUNIT, t.scale))
		y_adjust = P_ReturnThrustY(t, t.angle, FixedMul(-12*FRACUNIT, t.scale))
	end

	local dist = P_AproxDistance(P_AproxDistance(t.x + x_adjust - actor.x, t.y + y_adjust - actor.y), t.z + z_adjust - actor.z)
	local speed = dist / actor.fuse

	actor.momx = FixedMul(FixedDiv(t.x + x_adjust - actor.x, dist), speed)
	actor.momy = FixedMul(FixedDiv(t.y + y_adjust - actor.y, dist), speed)
	actor.momz = FixedMul(FixedDiv(t.z + z_adjust - actor.z, dist), speed)
end

function A_LightCircleThink(actor, var1, var2) --Thinker for the lightning circle
	local t = actor.target
	local p = t.player

	if not (t.state >= S_PLAY_SUPER_TRANS1 and t.state <= S_PLAY_SUPER_TRANS6)
		P_RemoveMobj(actor)
		return
	end

	actor.scale = t.scale

	local fa = FixedAngle(actor.info.speed) + actor.angle - ANGLE_90
	actor.angle = fa + ANGLE_90

	actor.circleradius = $ + var1
	local radius = FixedMul(actor.circleradius, actor.scale)

	local offsetx = FixedMul(cos(fa), radius)
	local offsety = FixedMul(sin(fa), radius)
	local offsetz = 0

	if t.eflags & MFE_VERTICALFLIP
		actor.eflags = $ | MFE_VERTICALFLIP
	end

	P_TeleportMove(actor, t.x + offsetx, t.y + offsety, t.z + offsetz)
end

//Reset stuff when the Player spawns
addHook("PlayerSpawn", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == "jetmetalsonic") return end
	ResetAbilities(p)
	ResetCharge(p)
	p.combocount = 0
	if p.deadalready
		p.deadalready = nil
	end
end)

//Recurl by pressing JUMP or SPIN while in Spring state
addHook("SpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end
	if mo.state == S_PLAY_SPRING
	and not (p.pflags & PF_SPINDOWN)
		p.pflags = $|PF_JUMPED|PF_SPINDOWN
		mo.state = S_PLAY_ROLL
		mo.eflags = $ & ~MFE_SPRUNG
		S_StartSound(mo, sfx_3db16)
		return true
	end
end)
addHook("JumpSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if mo.state == S_PLAY_SPRING
	and not (p.pflags & PF_JUMPDOWN)
		p.pflags = $|PF_JUMPED|PF_JUMPDOWN
		mo.state = S_PLAY_ROLL
		mo.eflags = $ & ~MFE_SPRUNG
		S_StartSound(mo, sfx_3db16)
	end

	if mo.state == S_PLAY_DASHREADY
		mo.state = S_PLAY_ROLL
	end
end)

//Hook charging to double jumping
addHook("AbilitySpecial", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
	if p.powers[pw_carry] return end
	if not p.charging
	and not (p.pflags & (PF_SPINDOWN|PF_JUMPDOWN))
		if p.charges > 0

			--Start Charging
			S_StartSound(p.mo, sfx_cdfm44)
			p.chrgWBounce = 0
			p.chrgGBounce = false
			p.charging = true
			if p.chrgStomping
				p.chrgStomping = false
				p.prevChrgStomp = true
			end
			p.storeSpd = FixedHypot(p.rmomx, p.rmomy)
			p.pflags = ($|PF_NOJUMPDAMAGE) & ~PF_SPINNING
			p.ghostTrail = 0
		else
			
		end
	end
	return true
end)

//Hook Homing Thok to SPIN while jumping
addHook("JumpSpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end

	--Perform the Homing Thok
	if not p.Homing
	and not (p.pflags & (PF_SPINDOWN|PF_THOKKED))
	and mo.tracer and mo.tracer.valid and mo.tracer.health
		S_StartSound(mo, sfx_beflap)
		mo.angle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)
		SpawnLaunchEffect(p, 1, 3*FRACUNIT/5, p.effcolor)
		p.Homing = HOME_TIMER
		p.pflags = $ & ~PF_NOJUMPDAMAGE
		p.chrgStomping = false
		p.ghostTrail = 0
		return true
	end
end)

//Hook Charging behaviour and follow-up buttons in the PreThinkFrame
addHook("PreThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
		if p.charges == nil
			p.charges = 0
		end

		if p.powers[pw_carry]
			ResetCharge(p)
		end

		--Set visual effect colour
		if p.menu.contents[1][3].value == 0
			p.effcolor = SKINCOLOR_YELLOW
		elseif p.menu.contents[1][3].value == 1
			p.effcolor = SKINCOLOR_ORANGE
		elseif p.menu.contents[1][3].value == 2
			p.effcolor = SKINCOLOR_CRIMSON
		elseif p.menu.contents[1][3].value == 3
			p.effcolor = SKINCOLOR_AETHER
		end

		if p.charging
		and not P_IsObjectOnGround(mo)

			--Lose charging state and all charges when hurt
			if P_PlayerInPain(p)
				ResetCharge(p)
				p.charges = 0
			end

			--Lose charging state when hitting a spring
			if mo.eflags & MFE_SPRUNG
				ResetCharge(p)
			end

			--Flash colours if not super
			if not p.powers[pw_super]
				if not p.prevcolor
					p.prevcolor = mo.color
				end
				if leveltime % 4 == 0
					mo.color = p.effcolor
				elseif leveltime % 4 == 2
					mo.color = p.prevcolor
				end
			end
			if leveltime % 4 == 0
				mo.colorized = true
			elseif leveltime % 4 == 2
				mo.colorized = false
			end

			--If the JUMP button is held then increase charge
			if (p.cmd.buttons & BT_JUMP)
			and p.chrgbuildUp < MAX_CHARGE_TIME
			and not P_PlayerInPain(p)
				p.chrgbuildUp = $+1

				--Set charging animation
				mo.state = S_PLAY_WALK
				mo.sprite2 = SPR2_CHRG
				mo.frame = A

				--Charge 50 percent faster when super or when having speed shoes
				if (p.powers[pw_super] or p.powers[pw_sneakers])
				and leveltime % 2 == 0
				and p.chrgbuildUp < MAX_CHARGE_TIME
					p.chrgbuildUp = $+1
				end
				
				--Charge 25% faster while Dash Mode is active
				if p.dashMode >= DASHMODE_ON
				and leveltime % 4 == 0
				and p.chrgbuildUp < MAX_CHARGE_TIME
					p.chrgbuildUp = $+1
				end

				--Reduce momentum and fall slowly while charging
				mo.momx = 8*$/9
				mo.momy = 8*$/9
				mo.momz = 2*$/5

				--Audio and visual cues for charge levels
				if p.chrgbuildUp >= HALF_CHARGE_BUILDUP
				and not p.halfchargeeff
					S_StartSound(mo, sfx_s3ka8)
					SpawnChargeEffect(p, p.effcolor)
					p.halfchargeeff = true
				elseif p.chrgbuildUp >= FULL_CHARGE_BUILDUP
				and not p.fullchargeeff
					S_StartSound(mo, sfx_s3ka8)
					SpawnChargeEffect(p, p.effcolor)
					p.fullchargeeff = true
					p.prepchrgBounce = true --Bounce possible
				end

				--If the SPIN button is pressed while charging, perform the Charge Jump / Stomp
				if (p.cmd.buttons & BT_SPIN)
				and not (p.pflags & PF_SPINDOWN)
					p.pflags = $|PF_SPINDOWN
					if p.menu.contents[1][5].value == 0
						if (not p.chrgjumpUsed or (p.powers[pw_super] and not p.superchrgjumpUsed))
						and p.chrgbuildUp < MAX_CHARGE_TIME
							S_StartSound(mo, sfx_beflap)
							ConvertCharge(p)
							SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)
							doChargeJump(p)
						else
							mo.state = S_PLAY_FALL
							ResetCharge(p)
							p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
						end
					else
						S_StartSound(mo, sfx_beflap)
						ConvertCharge(p)
						SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)
						doChargeStomp(p)
					end
				end

			--If the JUMP button is released while charging, perform the Charge Thok or Charge Stomp / Jump
			elseif not (p.cmd.buttons & BT_JUMP)
			and p.chrgbuildUp < MAX_CHARGE_TIME

				--If holding any direction then perform Charge Thok
				if P_GetPlayerControlDirection(p) ~= 0
					S_StartSound(mo, sfx_beflap)
					ConvertCharge(p)
					SpawnLaunchEffect(p, 1, p.effsize, p.effcolor)
					doChargeThok(p)

				--Otherwise, perform Charge Stomp / Jump
				else
					if p.menu.contents[1][5].value == 0
						S_StartSound(mo, sfx_beflap)
						ConvertCharge(p)
						SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)
						doChargeStomp(p)
					else
						if (not p.chrgjumpUsed or (p.powers[pw_super] and not p.superchrgjumpUsed))
						and p.chrgbuildUp < MAX_CHARGE_TIME
							S_StartSound(mo, sfx_beflap)
							ConvertCharge(p)
							SpawnLaunchEffect(p, 0, p.effsize, p.effcolor)
							doChargeJump(p)
						else
							mo.state = S_PLAY_FALL
							ResetCharge(p)
							p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
						end
					end
				end

			--If charging longer than MAX_CHARGE_TIME cancel charge and lose all charges
			elseif not P_PlayerInPain(p)
				mo.state = S_PLAY_FALL
				ResetCharge(p)
				p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
				p.charges = 0
			end
		else
			S_StopSoundByID(mo, sfx_cdfm44)
			if p.prevcolor
				mo.color = p.prevcolor
				p.prevcolor = nil
			end
			mo.colorized = false
		end
	end
end)

//Jet boosters and aura effect hooked into ThinkFrames
addHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if mo and mo.valid and mo.skin == SKIN

			--Spawn boosters if they do not exist yet
			if not p.deadalready
				if not p.jetmetbstrA or not p.jetmetbstrA.valid
					p.jetmetbstrA = P_SpawnMobj(mo.x, mo.y, mo.z, MT_JETMETBSTERA)
					p.jetmetbstrA.target = mo
					p.jetmetbstrA.skin = mo.skin
				end
				if not p.jetmetbstrB or not p.jetmetbstrB.valid
					p.jetmetbstrB = P_SpawnMobj(mo.x, mo.y, mo.z, MT_JETMETBSTERB)
					p.jetmetbstrB.target = mo
					p.jetmetbstrB.skin = mo.skin
				end
			end
			if p.jetmetbstrA and p.jetmetbstrA.valid
			and p.jetmetbstrB and p.jetmetbstrB.valid
				p.jetmetbstrA.sprite2 = SPR2_JMBA
				p.jetmetbstrB.sprite2 = SPR2_JMBB

				--Boosters crackle with electricity when super
				if p.powers[pw_super]
					JetScreamEffect(p.jetmetbstrA, p.effcolor, 3, 4, MT_JETMETZAP)
					JetScreamEffect(p.jetmetbstrB, p.effcolor, 3, 4, MT_JETMETZAP)
				end
			end

			--Spawn jetfumes under specific circumstances
			if (mo.state == S_PLAY_CHRGJMP or mo.state == S_PLAY_CHRGTHK or p.chrgStomping)
			and p.ghostTrail
				if not p.chrgStomping
					local x_adjust = P_ReturnThrustX(mo, mo.angle, FixedMul(-15*FRACUNIT, mo.scale))
					local y_adjust = P_ReturnThrustY(mo, mo.angle, FixedMul(-15*FRACUNIT, mo.scale))
					local z_adjust = mo.height/2
					if mo.state == S_PLAY_CHRGJMP
						z_adjust = FixedMul(12*FRACUNIT, mo.scale)
					end
					if mo.eflags & MFE_VERTICALFLIP
						z_adjust = -$
					end

					if not (mo.eflags & MFE_UNDERWATER)
						local jetfume = P_SpawnMobj(mo.x + x_adjust, mo.y + y_adjust, mo.z + z_adjust, MT_THOK)
						jetfume.fuse = 1
						jetfume.colorized = true
						jetfume.color = p.effcolor
						jetfume.sprite = SPR_JETF
						jetfume.scale = 2*mo.scale/3
						if mo.eflags & MFE_VERTICALFLIP
							jetfume.eflags = $|MFE_VERTICALFLIP
						end

						if leveltime % 3 == 0
							jetfume.frame = A|FF_FULLBRIGHT
						elseif leveltime % 3 == 1
							jetfume.frame = B|FF_FULLBRIGHT
						else
							jetfume.frame = C|FF_FULLBRIGHT
						end
						P_SpawnGhostMobj(jetfume)

					--if underwater, spawn bubbles instead
					else
						P_SpawnMobj(mo.x+x_adjust, mo.y+y_adjust, mo.z+z_adjust, MT_SMALLBUBBLE)
					end
				end
				BoosterJetFume(p.jetmetbstrA, true)
				BoosterJetFume(p.jetmetbstrB, true)
			end
			if mo.state == S_PLAY_RUN
			or mo.state == S_PLAY_DASH
				if p.Homing
					BoosterJetFume(p.jetmetbstrA, true)
					BoosterJetFume(p.jetmetbstrB, true)
					p.followmobj.color = p.effcolor
					P_SpawnGhostMobj(p.followmobj)
				else
					local defnormspd = skins[mo.skin].normalspeed
					if p.dashMode >= DASHMODE_ON
					and ((leveltime % 8 == 0 and p.normalspeed < defnormspd + FixedMul(defnormspd, DASHSPD_MOD_CAP/2))
					or (leveltime % 4 == 0 and p.normalspeed >= defnormspd + FixedMul(defnormspd, DASHSPD_MOD_CAP/2)))
							BoosterJetFume(p.jetmetbstrA, true)
							BoosterJetFume(p.jetmetbstrB, true)
					else
						BoosterJetFume(p.jetmetbstrA, false)
						BoosterJetFume(p.jetmetbstrB, false)
					end
				end
			end

			--Use aura effect under certain circumstances
			if p.Homing
			or p.chrgStomping
			or mo.state == S_PLAY_CHRGJMP
			or mo.state == S_PLAY_CHRGTHK
				local aura = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				aura.fuse = 1
				aura.angle = mo.angle
				aura.sprite = SPR_JMSA
				aura.color = p.effcolor
				aura.dispoffset = 1
				if not (aura.flags2 & MF2_LINKDRAW)
					aura.flags2 = $|MF2_LINKDRAW
				end
				if mo.eflags & MFE_VERTICALFLIP
					aura.eflags = $|MFE_VERTICALFLIP
					P_TeleportMove(aura, aura.x, aura.y, aura.z - FixedMul(12*FRACUNIT, mo.scale))
				end

				if p.Homing
				or mo.state == S_PLAY_CHRGTHK
					aura.frame = A
				elseif p.chrgStomping
					aura.frame = B
				else
					aura.frame = C
				end

				if leveltime % 2 == 1
					aura.frame = $|FF_TRANS50|FF_FULLBRIGHT
				else
					aura.frame = $|FF_TRANS70|FF_FULLBRIGHT
				end
			end

			--Secondary charge effect
			if p.chrgbuildUp > 0 and p.chrgbuildUp < HALF_CHARGE_BUILDUP
				SpawnChargeEffect2(mo, p.effcolor, 3)
			elseif p.chrgbuildUp >= HALF_CHARGE_BUILDUP and p.chrgbuildUp < FULL_CHARGE_BUILDUP
				SpawnChargeEffect2(mo, p.effcolor, 1)
			end
		end
	end
end)

//Set followmobj color in PostThinkFrame
addHook("PostThinkFrame", do
	for p in players.iterate
		if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
		p.followmobj.colorized = true
		p.followmobj.color = p.effcolor
	end
end)

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if mo.health <= 0
		if p.chrgStomping
			mo.spritexscale = FRACUNIT
			mo.spriteyscale = FRACUNIT
		end
		p.dashMode = 0
		DashModeReset(p)
		if p.camerascale ~= skins[mo.skin].camerascale
			p.camerascale = skins[mo.skin].camerascale
		end
		--Custom death animations
		if mo.state == S_PLAY_DEAD
		or mo.state == S_PLAY_JMSDEAD
			mo.flags = $ & ~MF_NOCLIPHEIGHT

			if P_IsObjectOnGround(mo)
				mo.state = S_PLAY_JMSDEAD
				if p.deadalready
					mo.momx = 0
					mo.momy = 0
					mo.rollangle = 0
				end
			else
				mo.state = S_PLAY_DEAD
				mo.rollangle = $ + ANG10
			end

			if not p.deadalready
				P_InstaThrust(mo, mo.angle, p.prevspeed/2)
				mo.momz = 2*$/3
				p.deadalready = true
			end
		end
		return
	end
	p.prevspeed = FixedHypot(mo.momx, mo.momy)

	JetScreamEffect(mo, p.effcolor, 1, P_RandomRange(1, 8), MT_JETMETZAP)
	if p.powers[pw_super]
		JetScreamEffect(mo, p.effcolor, 3, 8, MT_JETMETZAP)
		if mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6
			for i = 1, 4
				JetScreamEffect(mo, p.effcolor, 1, 8, MT_JETMETZAP)
			end
		end
	end

	if (maptol & TOL_NIGHTS) return end
	if p.powers[pw_carry]
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
		end
		return
	end

	--Go super by pressing the super button, adapted and modified from Reveries by Katmint and MaxControl by Inferno
	local superbutton = BT_CUSTOM1
	if p.menu.contents[1][6].value == 1
		superbutton = BT_CUSTOM2
	elseif p.menu.contents[1][6].value == 2
		superbutton = BT_CUSTOM3
	elseif p.menu.contents[1][6].value == 3
		superbutton = BT_TOSSFLAG
	end
	if All7Emeralds(emeralds) and (p.rings > 49) and (p.playerstate == PST_LIVE)
	and not p.powers[pw_super]
	and not (p.pflags & PF_SPINNING) and not (p.pflags & PF_THOKKED)
	and not (p.panim == PA_ABILITY) and not (p.panim == PA_ABILITY2) and not (p.panim == PA_PAIN) and not (p.panim == PA_ETC)
	and not (mo.state == S_PLAY_SPRING) and not (mo.state == S_PLAY_FALL)
	and not p.chrgStomping and not p.Homing and not p.charging
	and not P_IsObjectOnGround(mo)
		if p.cmd.buttons & superbutton
			p.charflags = $|SF_SUPER
			P_DoSuperTransformation(p)
			P_StartQuake(20*FRACUNIT, TICRATE/2)
			SpawnLightCircle(p, p.effcolor)
		end
	elseif (p.charflags & SF_SUPER)
	and not p.powers[pw_super]
		p.charflags = $ & ~SF_SUPER
	end

	--Do not air walk or air stand
	if p.pflags & ~(PF_JUMPED|PF_SPINNING)
	and not P_PlayerInPain(p)
	and not P_IsObjectOnGround(mo)
	and not p.powers[pw_nocontrol]
	and not p.charging
		if mo.state == S_PLAY_STND
		or mo.state == S_PLAY_WAIT
		or mo.state == S_PLAY_WALK
		or mo.state == S_PLAY_RUN
		or (mo.state == S_PLAY_DASH and not p.Homing)
			mo.state = S_PLAY_SPRING
		end
	end

	--Create trail when required
	if not p.ghostTrail
		p.ghostTrail = 0
	end
	if p.ghostTrail > 0
	or (P_IsObjectOnGround(mo) and p.pflags & PF_SPINNING and not (p.pflags & PF_STARTDASH))
		local playerghost = P_SpawnGhostMobj(mo)
		local boosterghostA = P_SpawnGhostMobj(p.jetmetbstrA)
		local boosterghostB = P_SpawnGhostMobj(p.jetmetbstrB)
		playerghost.colorized = true
		playerghost.color = p.effcolor
		boosterghostA.colorized = true
		boosterghostA.color = p.effcolor
		boosterghostB.colorized = true
		boosterghostB.color = p.effcolor
		if p.ghostTrail > 0
			p.ghostTrail = $-1
		end
	end

	//CHARGE ABILITIES
	p.thokspeedcap = 3*p.normalspeed

	--Return jump damage when entering a rolling state in the air (used for wall bounce)
	if mo.state == S_PLAY_ROLL
	and not P_IsObjectOnGround(mo)
		p.pflags = $ & ~PF_NOJUMPDAMAGE
	end

	--Return to rolling state at the height of a Charge Jump
	if mo.state == S_PLAY_CHRGJMP
		if (mo.momz < 5*FRACUNIT and not (mo.eflags & MFE_VERTICALFLIP))
		or (mo.momz > -5*FRACUNIT and mo.eflags & MFE_VERTICALFLIP)
			mo.tics = 1
		end
	end

	--Disable stomp and bounce behaviour when hitting a spring, entering goo or getting hurt
	if (mo.eflags == MFE_SPRUNG or mo.state == S_PLAY_SPRING or mo.eflags & MFE_GOOWATER or P_PlayerInPain(p))
	and (p.chrgStomping or p.chrgWBounce)
		if p.chrgWBounce
			p.chrgWBounce = 0
		end
		if p.chrgGBounce
			p.chrgGBounce = false
			if mo.eflags == MFE_SPRUNG
			or mo.state == S_PLAY_SPRING
				P_SetObjectMomZ(mo, 3*mo.momz/4, true) --Increase Z-momentum by 75% when hitting a spring with a stomp
			end
		end
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
		end
	end

	//This next bit is to make comboing and ground bouncing play nice with each other
	if not p.chrgGBouncing
		p.chrgGBouncing = false
	end
	if p.chrgGBouncing
	and not p.chrgGBounce
		p.chrgGBouncing = false
	end

	--Reset everything air-related when on the ground and not bouncing
	if (P_IsObjectOnGround(mo) or (mo.eflags & MFE_JUSTHITFLOOR))
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
		end
		if not p.chrgGBounce
			p.chrgGBounce = false
			ResetCharge(p)
			ResetAbilities(p)
			if not p.powers[pw_super]
				p.charges = MAX_CHARGES
			else
				p.charges = SUPER_MAX_CHARGES
			end
			p.ghostTrail = 0
			p.pflags = $ & ~PF_NOJUMPDAMAGE

		--Ground bounce behaviour
		else
			p.chrgGBounce = false
			mo.state = S_PLAY_JUMP
			p.pflags = ($ & ~PF_SPINNING)|PF_JUMPED
			FloorBreak(mo, mo.subsector.sector)

			-- Ground bounce dust from Max Control Pack by Inferno
			for i = 0, 8
				local fa = (i*(ANGLE_180/4))
				local dust = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
				dust.momx = FixedMul(sin(fa),20*FRACUNIT)
				dust.momy = FixedMul(cos(fa),20*FRACUNIT)
			end

			local GBOUNCE_HEIGHT = BASE_GBOUNCE_HEIGHT
			if p.powers[pw_super]
				GBOUNCE_HEIGHT = 6*$/5 --Increase modifier by 20 percent when Super
			end
			if mo.eflags & MFE_UNDERWATER
				GBOUNCE_HEIGHT = 3*$/5 --Decrease modifier by 40 percent when underwater
			end
			P_SetObjectMomZ(mo, GBOUNCE_HEIGHT, false)
			p.ghostTrail = 14
			S_StartSound(mo, sfx_mspogo)
		end
	end
	

	if p.chrgWBounce > 0
		p.chrgWBounce = $-1
	end

	//HOMING THOK
	--Perform sprite scaling and keep Z-momentum constant while stomping
	if p.chrgStomping
		P_SetObjectMomZ(mo, p.STOMPSPEED, true)
		if P_PlayerInPain(p)
		or mo.eflags & MFE_GOOWATER
		or mo.state ~= S_PLAY_ROLL
			p.chrgStomping = false
		end

		if mo.spriteyscale < 3*FRACUNIT/2
			mo.spriteyscale = $+FRACUNIT/10
		end
		if mo.spritexscale > 9*FRACUNIT/10
			mo.spritexscale = $-FRACUNIT/10
		end

	elseif p.prevChrgStomp
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
		p.prevChrgStomp = false
	end

	if not p.Homing
	and not (p.pflags & PF_THOKKED)
	and not P_IsObjectOnGround(mo)
	and not P_PlayerInPain(p)
		mo.tracer = JetLookForEnemies(p, SEARCH_DIST, SEARCH_ANGLE, true, MF_SPRING|MF_MONITOR) --Search for valid targets
		if mo.tracer and mo.tracer.valid and mo.tracer.health
		and not (mo.state == S_PLAY_SPRING or mo.state == S_PLAY_FALL)
		and not p.charging
			P_SpawnLockOn(p, mo.tracer, S_LOCKON2) --Spawn LockOn object
		end
	end

	--Perform homing attack
	if P_PlayerInPain(p)
	or (mo.eflags & MFE_SPRUNG)
		p.Homing = 0
	end
	if p.Homing
		if not (mo.tracer and mo.tracer.valid and mo.tracer.health)
			p.Homing = 0
		end

		if not p.Homing
			mo.state = S_PLAY_FALL
			p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
		else
			mo.state = S_PLAY_DASH
			p.pflags = ($ & ~PF_NOJUMPDAMAGE)|PF_STASIS
			JetHoming(mo, mo.tracer, SPEED_MOD)
			p.Homing = $-1
		end
	end

	//DASH MODE
	local speed = FixedHypot(mo.momx, mo.momy)
	local runspeed = FixedMul(FixedMul(p.runspeed, mo.friction), mo.scale)
	local defnormspd = skins[mo.skin].normalspeed
	local normspd = FixedMul(FixedMul(defnormspd, mo.friction), mo.scale)

	if not p.dashMode
		p.dashMode = 0
	end

	--Build up dashMode when moving at high enough speed or charging a spindash
	if speed >= runspeed
	or p.pflags & PF_STARTDASH
		if p.dashMode < DASHMODE_CAP
			p.dashMode = $+1
			if (p.pflags & PF_STARTDASH or speed >= normspd-FixedMul(3*FRACUNIT, mo.scale))
			and leveltime % 2 == 0
			and p.dashMode < DASHMODE_CAP
				p.dashMode = $+1
			end
			if p.dashMode >= DASHMODE_ON/2 and p.dashMode <= DASHMODE_ON
				SpawnChargeEffect2(mo, p.effcolor, 2)
			end
		end

	--Lose dashMode buildup under certain circumstances
	elseif P_IsObjectOnGround(mo)
	and not p.chrgGBounce --Synergy with Charge Abilities
		p.dashMode = $-DASHMODE_DECAY
		if p.dashMode < 0
			p.dashMode = 0
		end
	end

	--Fake super peelout based on commented out code from CrossMomentum by Frostiikin
	if p.pflags & PF_STARTDASH
	and p.dashMode >= DASHMODE_ON
		mo.state = S_PLAY_DASHREADY
	end
	if p.statelast
		if (mo.state == S_PLAY_ROLL)
		and (p.statelast == S_PLAY_DASHREADY)
		and P_IsObjectOnGround(mo)
			mo.state = S_PLAY_RUN
			p.pflags = $ & ~PF_SPINNING
			SpawnLaunchEffect(p, 1, FRACUNIT, p.effcolor)
			P_StartQuake(20*FRACUNIT, TICRATE/5)
		end
	end
	p.statelast = mo.state

	--Do various things once Dash Mode is built up high enough
	if p.dashMode >= DASHMODE_ON
		if not p.dashModeactivate
			S_StartSound(mo, sfx_cdfm40) --Vanilla Dash Mode activation sound
			p.jumpfactor = 11*$/10 --Increase jump factor by 10%
			if not (p.pflags & PF_STARTDASH)
				SpawnLaunchEffect(p, 1, FRACUNIT, p.effcolor)
				P_StartQuake(20*FRACUNIT, TICRATE/5)
			end
			p.dashModeactivate = true
		end
		if not p.chrgStomping
			p.charflags = $|SF_RUNONWATER --Run on water
		else
			if not (skins[p.mo.skin].flags & SF_RUNONWATER)
				p.charflags = $ & ~SF_RUNONWATER --Disable water running while stomping
			end
		end

		--Spawn pulsing effect
		if (leveltime % 8 == 0 and p.normalspeed < defnormspd + FixedMul(defnormspd, DASHSPD_MOD_CAP/2))
		or (leveltime % 4 == 0 and p.normalspeed >= defnormspd + FixedMul(defnormspd, DASHSPD_MOD_CAP/2))
			p.followmobj.color = p.effcolor
			local dashghost = P_SpawnGhostMobj(mo)
			dashghost.tracer = mo
			dashghost.colorized = true
			dashghost.blendmode = AST_ADD
			dashghost.scale = mo.scale
			dashghost.destscale = 4*mo.scale/3
			dashghost.scalespeed = abs(dashghost.scale - dashghost.destscale) / dashghost.fuse
			dashghost.dashactive = true
		end

		if p.normalspeed < FixedMul(defnormspd, DASHSPD_MOD_CAP)
		and P_IsObjectOnGround(mo)
		and speed >= normspd
			p.normalspeed = $ + defnormspd / (3*TICRATE) --Gradually increase normalspeed
			if p.normalspeed > FixedMul(defnormspd, DASHSPD_MOD_CAP)
				p.normalspeed = FixedMul(defnormspd, DASHSPD_MOD_CAP) --Do not overshoot cap
			end
		end

		if speed > defnormspd
			if not p.dashcamera
				p.camerascale = 23*$/20 --Zoom out camera by 15%
				p.dashcamera = true
			end
		else

			--Reset things when at lower speed
			if p.dashcamera
				p.camerascale = skins[mo.skin].camerascale
				p.dashcamera = false
			end
		end
	else
		DashModeReset(p)
	end
	if P_IsObjectOnGround(mo)
	and mo.state == S_PLAY_RUN
	and speed > defnormspd
		mo.state = S_PLAY_DASH
	elseif P_IsObjectOnGround(mo)
	and mo.state == S_PLAY_DASH
	and speed < defnormspd
	and not p.Homing
		mo.state = S_PLAY_RUN
	end

	//COMBO COUNTER
	if P_IsObjectOnGround(p.mo)
	and not p.chrgGBouncing
		p.combocount = 0 --Reset combo counter when on the ground and not bouncing
	end
end)

addHook("PlayerCanDamage", function(p) --Damage enemies while dashing
	if not p.dashMode
		p.dashMode = 0
	end
	if p.dashMode >= DASHMODE_ON return true end
end)

//Behaviour for when an enemy is damaged
addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	local p = i.player
	if not P_IsObjectOnGround(i)
	and t.valid
		if p.chrgStomping
			p.chrgStomping = false
			p.prevChrgStomp = true
			i.momz = $/2
		end
		if t.flags & MF_ENEMY|MF_BOSS

			--Recover a single charge if the player has less than one
			if p.charges < 1
				p.charges = 1
			end
			ResetAbilities(p)
			
			p.combocount = $+1 --Increase combo counter when hitting an enemy or boss while airborne
		end
	end
	if t.valid
	and (t.flags & MF_ENEMY|MF_BOSS|MF_MONITOR)
	and p.Homing
		local speed = p.actionspd
		if i.eflags & MFE_UNDERWATER
			speed = 3*$/5 --Reduce speed by 40% when underwater
		end
		P_InstaThrust(i, i.angle, FixedMul(speed/6, i.scale))
		P_SetObjectMomZ(i, speed/10, false)
		p.pflags = $ & ~PF_THOKKED
		i.state = S_PLAY_POSTMOVE
		p.Homing = 0
	end
end)

//Line detection code from Krabs's Rebound Dash
addHook("MobjLineCollide", function(mo, line)
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end
	local p = mo.player

	local side = P_PointOnLineSide(mo.x, mo.y, line)
	local sector = nil

	if side == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end

	if sector == nil return end

	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BLOCKPLAYER)
			continue
		end

		local topheight = rover.topheight
		local bottomheight = rover.bottomheight

		if (rover.t_slope)
			topheight = P_GetZAt(rover.t_slope, mo.x, mo.y)
		end
		if (rover.b_slope)
			bottomheight = P_GetZAt(rover.b_slope, mo.x, mo.y)
		end

		if (topheight < mo.z)
			continue
		end

		if (bottomheight > mo.z + mo.height)
			continue
		end

		if (rover.flags & FF_BUSTUP) //and not (rover.flags & FF_STRONGBUST)
			p.bustsector = sector
			p.bustrover = rover
		end
		break
	end
end, MT_PLAYER)

//Behaviour when movement is blocked
addHook("MobjMoveBlocked", function(mo)
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
	local p = mo.player

	--Bounce when hitting a wall with a fully charged Charge Thok
	if p.chrgWBounce
		P_BounceMove(mo)
		mo.momx = FixedMul($, WALLBOUNCEXY_FACTOR)
		mo.momy = FixedMul($, WALLBOUNCEXY_FACTOR)
		P_SetObjectMomZ(mo, WALLBOUNCEZ_FACTOR, false)

		S_StartSound(mo, sfx_mspogo)
		p.chrgWBounce = 0
		p.ghostTrail = 14
		mo.state = S_PLAY_POSTMOVE
		p.pflags = $|PF_NOJUMPDAMAGE

		WallBounceDust(mo)

		--Break bustable FOFs from Krabs's Rebound Dash
		if p.bustsector and p.bustrover and p.bustsector.valid and p.bustrover.valid
			EV_CrumbleChain(p.bustsector, p.bustrover)
			p.bustsector = nil
			p.bustrover = nil

			--Recover a single charge if the player has less than one
			if p.charges < 1
				p.charges = 1
			end
		end
	end
	
	if p.Homing
		p.Homing = 0
		mo.state = S_PLAY_FALL
		p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
	end
end, MT_PLAYER)

addHook("MobjMoveCollide", function(mo, mobj)
    if not (mo.valid and mo.skin == SKIN) return end
	local p = mo.player

	--Break spikes and saloon doors with fully charged thoks, modified from spike breaking code by ARJr
	if p.chrgWBounce
		if mobj.type == MT_SPIKE or mobj.type == MT_WALLSPIKE or mobj.type == MT_SALOONDOOR or mobj.type == MT_SALOONDOORCENTER
			if (mobj.z >= (mo.z - mo.height/5 - 16*FRACUNIT)) and (mobj.z <= (mo.z + 6*mo.height/5 + 16*FRACUNIT))
			and mobj.type == MT_SPIKE
				P_KillMobj(mobj, mo, mo)
			elseif (mobj.flags2 & MF2_OBJECTFLIP)
			and (mobj.z <= (mo.z - mo.height/5 - 16*FRACUNIT)) and (mobj.z >= (mo.z + 6*mo.height/5 + 16*FRACUNIT))
			and mobj.type == MT_SPIKE
				P_KillMobj(mobj, mo, mo)
			end

			--Destroy saloon doors
			if mobj.type == MT_SALOONDOORCENTER

				--Bounce off of the door
				local bouncex = P_ReturnThrustX(mo, mo.angle + ANGLE_180, FixedMul(FixedHypot(mo.momx, mo.momy), WALLBOUNCEXY_FACTOR))
				local bouncey = P_ReturnThrustY(mo, mo.angle + ANGLE_180, FixedMul(FixedHypot(mo.momx, mo.momy), WALLBOUNCEXY_FACTOR))
				mo.momx = bouncex
				mo.momy = bouncey
				P_SetObjectMomZ(mo, WALLBOUNCEZ_FACTOR, false)
				S_StartSound(mo, sfx_mspogo)

				p.chrgWBounce = 0
				p.ghostTrail = 14
				mo.state = S_PLAY_POSTMOVE
				p.pflags = $|PF_NOJUMPDAMAGE

				WallBounceDust(mo)

				--Debris code is modified from the ground bounce dust from Max Control Pack by Inferno 
				for i = 0, 8
					local fa = (i*(ANGLE_180/4))
					local debris = P_SpawnMobjFromMobj(mobj, 0, 0, mobj.info.height/2, MT_WOODDEBRIS)
					debris.momx = FixedMul(sin(fa), 14*FRACUNIT)
					debris.momy = FixedMul(cos(fa), 14*FRACUNIT)
				end
				S_StartSound(mobj, sfx_wbreak)

				P_KillMobj(mobj, mo, mo) -- Kill door
				
				--Recover a single charge if the player has less than one
				if p.charges < 1
					p.charges = 1
				end
			elseif mobj.type == MT_WALLSPIKE or mobj.type == MT_SALOONDOOR
				P_KillMobj(mobj, mo, mo)
			end
		end
	end

	--Break floor and ceiling spikes with fully charged stomps
	if p.chrgGBounce
	and mobj.type == MT_SPIKE
		if (mobj.z >= (mo.z - 16*FRACUNIT)) and (mobj.z <= (mo.z + mo.height + 16*FRACUNIT))
			P_KillMobj(mobj, mo, mo)
		elseif (mobj.flags2 & MF2_OBJECTFLIP)
		and (mobj.z <= (mo.z - 16*FRACUNIT)) and (mobj.z >= (mo.z + mo.height + 16*FRACUNIT))
			P_KillMobj(mobj, mo, mo)
		end
	end

	--Destroy spikes while dashing
	if p.dashMode >= DASHMODE_ON
	and mobj.type == MT_SPIKE
		if mobj.z >= (mo.z-16*FRACUNIT)
		and mobj.z <= (mo.z+mo.height+16*FRACUNIT)
			P_KillMobj(mobj, mo)
		elseif (mobj.flags2 & MF2_OBJECTFLIP)
		and (mobj.z <= mo.z-16*FRACUNIT)
		and (mobj.z >= mo.z+mo.height+16*FRACUNIT)
			P_KillMobj(mobj, mo)
		end
	end
end, MT_PLAYER)

//Hook custom object properties into the MobjThinker
addHook("MobjThinker", function(mo)
	if not (mo.tracer and mo.tracer.valid) return end
	local t = mo.tracer
	local p = t.player

	if mo.dashactive
		local z_adjust = 0
		if t.eflags & MFE_VERTICALFLIP
			z_adjust = FixedMul(t.scale - mo.scale, P_GetPlayerHeight(p))
			if t.state == S_PLAY_ROLL
			or t.state == S_PLAY_DASHREADY
				z_adjust = $ - P_GetPlayerHeight(p)/3
			end
		end
		if p.prevcolor
			mo.color = p.prevcolor
		else
			mo.color = t.color
		end
		P_TeleportMove(mo, t.x, t.y, t.z + z_adjust)
	end
end)

hud.add(function(v, p)
	if p.mo and p.mo.skin == "jetmetalsonic"
	and p.powers[pw_carry] ~= CR_NIGHTSMODE

		//Draw the Charge Meter on the HUD, based on Inuyasha's Boss HUD code
		if p.menu.contents[1][1].value --Check to see if the Charge Meter is enabled

			local posx = 17
			local posy = 57

			local CHARGECAP = 99
			local CHARGEFILL = (p.chrgbuildUp*100/FULL_CHARGE_BUILDUP)*CHARGECAP/100

			--Set charge meter fill colour based on menu setting
			local fillcolor = p.mo.color
			if p.prevcolor
				fillcolor = p.prevcolor
			end
			if p.menu.contents[1][2].value == 1
				fillcolor = SKINCOLOR_BLUE
			elseif p.menu.contents[1][2].value == 2
				fillcolor = SKINCOLOR_ORANGE
			elseif p.menu.contents[1][2].value == 3
				fillcolor = SKINCOLOR_RED
			elseif p.menu.contents[1][2].value == 4
				fillcolor = SKINCOLOR_ROSY
			elseif p.menu.contents[1][2].value == 5
				fillcolor = SKINCOLOR_PURPLE
			elseif p.menu.contents[1][2].value == 6
				fillcolor = SKINCOLOR_YELLOW
			elseif p.menu.contents[1][2].value == 7
				fillcolor = SKINCOLOR_GREEN
			end

			--Draw the charge meter
			posy = $+(CHARGECAP-1)
			for i = 1, CHARGECAP
				if CHARGEFILL >= i
					v.draw(posx, posy, v.cachePatch("CMTRFILL"), V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_PERPLAYER, v.getColormap(SKIN, fillcolor))
				else
					v.draw(posx, posy, v.cachePatch("CMTRBCKG"), V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_PERPLAYER)
				end
				posy = $-1
			end
			posy = $+1
			v.draw(posx, posy, v.cachePatch("CMTRBASE"), V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_PERPLAYER)
			posx = $+3
			posy = $+CHARGECAP+2
			if p.charges > 3
				v.drawString(posx, posy, "4", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, "left")
			elseif p.charges > 2
				v.drawString(posx, posy, "3", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, "left")
			elseif p.charges > 1
				v.drawString(posx, posy, "2", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, "left")
			elseif p.charges > 0
				v.drawString(posx, posy, "1", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, "left")
			else
				v.drawString(posx, posy, "0", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT, "left")
			end
		end

		//Draw the Combo Counter on the HUD
		if p.menu.contents[1][4].value --Check to see if the Combo Counter is enabled
			if p.combocount > 1
				local posx = 60
				local posy = 148

				--Shift the Combo Counter to the left if the Charge Meter is disabled
				if not p.menu.contents[1][1].value
					posx = $-18
				end

				--Counter cannot display beyond 2 digits
				if p.combocount > 99
					v.drawPaddedNum(posx, posy, 99, 2, V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
				else
					v.drawPaddedNum(posx, posy, p.combocount, 2, V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
				end

				posx = $-25
				posy = $+12
				v.drawString(posx, posy, "hits!", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_YELLOWMAP, "left")
			end
		end
	end
end, "game")